# TP 2 TS326

import numpy as np
import matplotlib.pyplot as plt


plt.close('all')

range_attributs = range(15,16) #range(2,16)
TauxLDA_kk = np.zeros((16))
TauxQDA_kk = np.zeros((16))
TauxNB_kk = np.zeros((16))
Taux1PPV_kk = np.zeros((16))
TauxKPPV_kk = np.zeros((16))

for NbAttributs in range_attributs: # Nombre d'attributs conservés

    # Lecture des données d'apprentissage
    XA = np.load('X_training.npy')
    YA = np.load('Y_training.npy')

    XA = XA[:, 0:NbAttributs]
    
    NA = XA.shape[0] #Nombre d'échantillons d'apprentissage
    Classes = np.unique(YA)
    Nb_classes = np.unique(YA).shape[0] #Nombre de classes
    
    #Récupération des données d'apprentissage pour chaque classe cl : tableaux XA_cl et YA_cl
    XA_cl = [] #1, Nb_classes);
    YA_cl = [] #cell(1, Nb_classes);
    NA_cl = np.zeros((Nb_classes))
    for cl in range(0, Nb_classes):
        pos = np.array(YA == Classes[cl])
        XA_cl.append(XA[np.squeeze(pos),:])
        YA_cl.append(YA[pos])
        NA_cl[cl] = np.sum(pos)
    
    #Tracé des formes dans l'espace des paramètres      
    symbole = ['rx', 'gx', 'bx', 'mx', 'kx', 'ro', 'go', 'bo', 'mo', 'ko']
    
    plt.figure(5)
    for cl in range(0, Nb_classes):
        plt.plot(XA_cl[cl][:,0], XA_cl[cl][:,1], symbole[cl])
    plt.xscale('log')
    plt.yscale('log')
    plt.legend("".join(str(x) for x in Classes))
    
    #%Calcul des paramètres pour les classifieurs LDA et QDA 
    
    mu_cl = [] # contenant les vecteurs moyens dans chaque groupe.
    C_cl = []  # contenant les matrices de covariance dans chaque groupe.
    
    for cl in range(0, Nb_classes): 
        # Calcul des moyennes dans chaque groupe
        mu_cl.append(np.mean(XA_cl[cl], axis=0))
        # Calcul des matrices de covariance dans chaque groupe
        C_cl.append(np.cov(XA_cl[cl].T))
    
    # Calcul de la matrice de covariance globale
    C = np.zeros((NbAttributs, NbAttributs))
    for cl in range(0, Nb_classes):
        C = C + NA_cl[cl] * C_cl[cl]
    C = C/NA
    
    #% Lecture des données de test                            
    
    XT = np.load('X_test.npy')
    YT = np.load('Y_test.npy')
    
    XT = XT[:, 0:NbAttributs]
    NT = XT.shape[0] # Nombre d'échantillons de test
    
    
    
    #%% Analyse Discriminante Linéaire
    
    print('\nAnalyse Discriminante Lin�aire\n')
    
    TauxLDA=0
    mat_conf_LDA = np.zeros((Nb_classes, Nb_classes))
    s=''
    
    G = np.zeros((NT,1)).astype('uint8')
    for i in range(0,NT):
        if (np.mod(i,100)==0):
            print('{}/{}'.format(i,NT))
            
        deltaL = np.zeros((Nb_classes))
        for cl in range(0,Nb_classes):
            #deltaL[cl] = np.dot(np.dot(XT[i,:],np.linalg.inv(C)),mu_cl[cl]) - 0.5*np.dot(np.dot(mu_cl[cl],np.linalg.inv(C)),mu_cl[cl]) + np.log(NA_cl[cl]/NA)
            deltaL[cl] = XT[i,:].T@np.linalg.inv(C)@mu_cl[cl] - 0.5*mu_cl[cl].T@np.linalg.inv(C)@mu_cl[cl] + np.log(NA_cl[cl]/NA)
    
        pos = np.argwhere(deltaL == np.max(deltaL))
        rnd = np.random.randint(pos.shape[0]) # tirage aléatoire parmi les ex-aequo.
        cl_predite = pos[rnd]
        
        if (cl_predite == YT[i]):
            TauxLDA = TauxLDA + 1
        
        G[i] = cl_predite
        
        mat_conf_LDA[YT[i], cl_predite] += 1
    
    TauxLDA /= NT
    print(TauxLDA)
    
    Taux2 = np.mean(G == YT)
    print(Taux2)
    
    #%% Analyse Discriminante Quadratique                     
    
    print('\nAnalyse Discriminante Quadratique\n')
    TauxQDA = 0
    mat_conf_QDA = np.zeros((Nb_classes, Nb_classes))
    s=''
    for i in range(0,NT):
        if (np.mod(i,100)==0):
            print('{}/{}'.format(i,NT))
        deltaQ = np.zeros((Nb_classes))
        
        for cl in range(0, Nb_classes):
            deltaQ[cl] = -0.5*np.log(np.linalg.det(C_cl[cl])) -0.5*np.dot(np.dot((XT[i,:]-mu_cl[cl]),np.linalg.inv(C_cl[cl])),(XT[i,:]-mu_cl[cl])) + np.log(NA_cl[cl]/NA)
         
        pos = np.argwhere(deltaQ == np.max(deltaQ))
        rnd = np.random.randint(pos.shape[0]) # tirage aléatoire parmi les ex-aequo.
        cl_predite = pos[rnd]
        
        if (cl_predite == YT[i]):
            TauxQDA = TauxQDA + 1
       
        mat_conf_QDA[YT[i], cl_predite] +=  1
    
    TauxQDA /= NT
    
    
    #%% Naïve Bayes                
    
    print('\nNaïve Bayes (Quadratic with only diagonal covariance matrix\n')
    TauxNB = 0
    mat_conf_NB = np.zeros((Nb_classes, Nb_classes))
    s=''
    for i in range(0,NT):
        if (np.mod(i,100)==0):
            print('{}/{}'.format(i,NT))
        deltaQ = np.zeros((Nb_classes))
        
        for cl in range(0, Nb_classes):
            deltaQ[cl] = -0.5*np.log(np.linalg.det(C_cl[cl]*np.eye(C.shape[0]))) -0.5*np.dot(np.dot((XT[i,:]-mu_cl[cl]),np.linalg.inv(C_cl[cl]*np.eye(C.shape[0]))),(XT[i,:]-mu_cl[cl])) + np.log(NA_cl[cl]/NA)
         
        pos = np.argwhere(deltaQ == np.max(deltaQ))
        rnd = np.random.randint(pos.shape[0]) # tirage aléatoire parmi les ex-aequo.
        cl_predite = pos[rnd]
        
        if (cl_predite == YT[i]):
            TauxNB = TauxNB + 1
       
        mat_conf_NB[YT[i], cl_predite] +=  1
    
    TauxNB /= NT
    
    
    #%% Classification K Plus Proches Voisins : K=1            
    
    print('\n1 Plus Proche Voisin\n')
    Taux1PPV = 0
    Decision = np.zeros((NT)).astype('int')
    mat_conf_1PPV = np.zeros((Nb_classes, Nb_classes))
    s=''
    for i in range(0,NT):
        if (np.mod(i,100)==0):
            print('{}/{}'.format(i,NT))
            
        # Distance Euclidienne
        d = np.sum((XT[i,:]-XA)**2,axis=1)
            
        jmin = np.argmin(d)
    
        Decision[i] = YA[jmin] 
        if (Decision[i] == YT[i]):
            Taux1PPV += 1
        
        mat_conf_1PPV[YT[i], Decision[i]] += 1
    
   
    Taux1PPV /= NT
    print(Taux1PPV)
    
    #%% Classification K Plus Proches Voisins :                
        
    print('\nK Plus Proches Voisins\n')
    K = 10
    TauxKPPV = 0
    mat_conf_KPPV = np.zeros((Nb_classes, Nb_classes))
    s=''
    
    for i in range(0,NT):
        if (np.mod(i,100)==0):
            print('{}/{}'.format(i,NT))
        
        d = np.zeros((NA,1))
        #     for j = 1:NA
        #         % Distance Euclidienne
        #         d(j,1) = sqrt((XT(i,:)-XA(j,:))*eye(NbAttributs)*(XT(i,:)-XA(j,:))');
        #         % Distance de Mahalanobis
        #         d(j,1) = (XT(i,:)-XA(j,:))*inv(C)*(XT(i,:)-XA(j,:))';
    
        #d = (XT(i,:)-XA)*inv(C)*(XT(i,:)-XA)';
    
        d = np.sum((XT[i,:]-XA)**2,axis=1)
        
        pos_kppv = np.argsort(d)
        Decide = YA[pos_kppv[0:K]] #étiquettes des K plus proches voisins
        
        a, tmp = np.histogram(Decide, range(0,Nb_classes+1))
        pos = np.argwhere(a == np.max(a))
        rnd = np.random.randint(len(pos))
        Decision[i] = pos[rnd]
        
        if (Decision[i]== YT[i]):
            TauxKPPV += 1
        
        mat_conf_KPPV[YT[i], Decision[i]] += 1
    
    TauxKPPV /= NT
    
    #% Affichage des résultats
    print(f"{TauxLDA = }")
    print(f"{TauxQDA = }")
    print(f"{TauxNB = }")
    print(f"{Taux1PPV = }")
    print(f"{TauxKPPV = }")
    
    #%
    TauxLDA_kk[NbAttributs] = TauxLDA
    TauxQDA_kk[NbAttributs] = TauxQDA
    TauxNB_kk[NbAttributs] = TauxNB
    Taux1PPV_kk[NbAttributs] = Taux1PPV
    TauxKPPV_kk[NbAttributs] = TauxKPPV
    
    ph_LDA = np.sum(np.sum(mat_conf_LDA,axis=0)*np.sum(mat_conf_LDA,axis=1)) / NT**2
    kappa_LDA = (TauxLDA - ph_LDA) / (1 - ph_LDA)
    
    ph_QDA = np.sum(np.sum(mat_conf_QDA,axis=0)*np.sum(mat_conf_QDA,axis=1)) / NT**2
    kappa_QDA = (TauxQDA - ph_QDA) / (1 - ph_QDA)
        
    ph_NB = np.sum(np.sum(mat_conf_NB,axis=0)*np.sum(mat_conf_NB,axis=1)) / NT**2
    kappa_NB = (TauxNB - ph_NB) / (1 - ph_NB)
    
    ph_1PPV = np.sum(np.sum(mat_conf_1PPV,axis=0)*np.sum(mat_conf_1PPV,axis=1)) / NT**2
    kappa_1PPV = (Taux1PPV - ph_1PPV) / (1 - ph_1PPV)
    
    ph_KPPV = np.sum(np.sum(mat_conf_KPPV,axis=0)*np.sum(mat_conf_KPPV,axis=1)) / NT**2
    kappa_KPPV = (TauxKPPV - ph_KPPV) / (1 - ph_KPPV)


#%%
if (len(range_attributs)>1):
    plt.figure()
    plt.plot(range_attributs,TauxLDA_kk[2:],'-*r')
    plt.plot(range_attributs,TauxQDA_kk[2:],'-*g')
    plt.plot(range_attributs,TauxNB_kk[2:],'-*b')
    plt.plot(range_attributs,Taux1PPV_kk[2:],'-*c')
    plt.plot(range_attributs,TauxKPPV_kk[2:],'-*k')
    plt.legend(['LDA', 'QDA', 'NB', '1PPV', 'K=10PPV'])
    plt.xlabel('Nb attributs (p)')
    plt.ylabel('OA')


#%% Classifiers with sklearn

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from sklearn.naive_bayes import GaussianNB

clf = LinearDiscriminantAnalysis()
clf.fit(XA, YA.ravel())
YT_pred = clf.predict(XT) 
accuracy_LDA_sklearn = np.mean(YT_pred==YT.ravel())
print(f"{accuracy_LDA_sklearn = }")

clf = QuadraticDiscriminantAnalysis()
clf.fit(XA, YA.ravel())
YT_pred = clf.predict(XT) 
accuracy_QDA_sklearn = np.mean(YT_pred==YT.ravel())
print(f"{accuracy_QDA_sklearn = }")

clf = GaussianNB()
clf.fit(XA, YA.ravel())
YT_pred = clf.predict(XT) 
accuracy_NB_sklearn = np.mean(YT_pred==YT.ravel())
print(f"{accuracy_NB_sklearn = }")

#PPV
from sklearn.neighbors import KNeighborsClassifier

clf = KNeighborsClassifier(n_neighbors=1)
clf.fit(XA, YA.ravel())
YT_pred = clf.predict(XT)
accuracy_1PPV_sklearn = np.mean(YT_pred==YT.ravel())
print(f"{accuracy_1PPV_sklearn = }")

clf = KNeighborsClassifier(n_neighbors=10)
clf.fit(XA, YA.ravel())
YT_pred = clf.predict(XT)
accuracy_KPPV_sklearn = np.mean(YT_pred==YT.ravel())
print(f"{accuracy_KPPV_sklearn = }")



